@@ -1,62 +1,144 @@
-# Copyright 2012 mdtraj developers
+##############################################################################
+# MDTraj: A Python Library for Loading, Saving, and Manipulating
+#         Molecular Dynamics Trajectories.
+# Copyright 2012-2013 Stanford University and the Authors
 #
-# This file is part of mdtraj
+# Authors: Peter Eastman, Robert McGibbon
+# Contributors:
 #
-# mdtraj is free software: you can redistribute it and/or modify it under the
-# terms of the GNU General Public License as published by the Free Software
-# Foundation, either version 3 of the License, or (at your option) any later
-# version.
+# MDTraj is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Lesser General Public License as
+# published by the Free Software Foundation, either version 2.1
+# of the License, or (at your option) any later version.
 #
-# mdtraj is distributed in the hope that it will be useful, but WITHOUT ANY
-# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
-# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
+# This library is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Lesser General Public License for more details.
 #
-# You should have received a copy of the GNU General Public License along with
-# mdtraj. If not, see http://www.gnu.org/licenses/.
-
-"""
-pdbfile.py: Used for loading PDB files.
-
-This is part of the OpenMM molecular simulation toolkit originating from
-Simbios, the NIH National Center for Physics-Based Simulation of
-Biological Structures at Stanford, funded under the NIH Roadmap for
-Medical Research, grant U54 GM072970. See https://simtk.org.
-
-Portions copyright (c) 2012 Stanford University and the Authors.
-Authors: Peter Eastman
-Contributors:
-
-Permission is hereby granted, free of charge, to any person obtaining a
-copy of this software and associated documentation files (the "Software"),
-to deal in the Software without restriction, including without limitation
-the rights to use, copy, modify, merge, publish, distribute, sublicense,
-and/or sell copies of the Software, and to permit persons to whom the
-Software is furnished to do so, subject to the following conditions:
-
-The above copyright notice and this permission notice shall be included in
-all copies or substantial portions of the Software.
-
-THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
-THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
-OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
-USE OR OTHER DEALINGS IN THE SOFTWARE.
-"""
-__author__ = "Peter Eastman"
-__version__ = "1.0"
+# You should have received a copy of the GNU Lesser General Public
+# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
+# Portions copyright (c) 2012 Stanford University and the Authors.
+#
+# Portions of this code originate from the OpenMM molecular simulation
+# toolkit. Those portions are Copyright 2008-2012 Stanford University
+# and Peter Eastman, and distributed under the following license:
+#
+# Permission is hereby granted, free of charge, to any person obtaining a
+# copy of this software and associated documentation files (the "Software"),
+# to deal in the Software without restriction, including without limitation
+# the rights to use, copy, modify, merge, publish, distribute, sublicense,
+# and/or sell copies of the Software, and to permit persons to whom the
+# Software is furnished to do so, subject to the following conditions:
+#
+# The above copyright notice and this permission notice shall be included in
+# all copies or substantial portions of the Software.
+#
+# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
+# THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+# USE OR OTHER DEALINGS IN THE SOFTWARE.
+##############################################################################
 
+from __future__ import print_function, division
 import os
 import numpy as np
 import xml.etree.ElementTree as etree
 from copy import copy
-from pdbstructure import PdbStructure
+from .pdbstructure import PdbStructure
 from mdtraj.topology import Topology
-from mdtraj.utils import ilen
-import element as elem
+from mdtraj.utils import ilen, cast_indices, convert
+from mdtraj.registry import _FormatRegistry
+from . import element as elem
+
+__all__ = ['load_pdb', 'PDBTrajectoryFile']
+
+##############################################################################
+# Code
+##############################################################################
+
+@_FormatRegistry.register_loader('.pdb')
+def load_pdb(filename, stride=None, atom_indices=None, frame=None):
+    """Load a RCSB Protein Data Bank file from disk.
+
+    Parameters
+    ----------
+    filename : str
+        Path to the PDB file on disk.
+    stride : int, default=None
+        Only read every stride-th model from the file
+    atom_indices : array_like, optional
+        If not none, then read only a subset of the atoms coordinates from the
+        file. These indices are zero-based (not 1 based, as used by the PDB
+        format). So if you want to load only the first atom in the file, you
+        would supply ``atom_indices = np.array([0])``.
+    frame : int, optional
+        Use this option to load only a single frame from a trajectory on disk.
+        If frame is None, the default, the entire trajectory will be loaded.
+        If supplied, ``stride`` will be ignored.
+
+    Returns
+    -------
+    trajectory : md.Trajectory
+        The resulting trajectory, as an md.Trajectory object.
+        
+    Examples
+    --------
+    >>> import mdtraj as md                                        # doctest: +SKIP
+    >>> pdb = md.load_pdb('2EQQ.pdb')                              # doctest: +SKIP
+    >>> print pdb                                                  # doctest: +SKIP
+    <mdtraj.Trajectory with 20 frames, 423 atoms at 0x110740a90>   # doctest: +SKIP
+
+    See Also
+    --------
+    mdtraj.PDBTrajectoryFile : Low level interface to PDB files
+    """
+    from mdtraj import Trajectory
+    if not isinstance(filename, str):
+        raise TypeError('filename must be of type string for load_pdb. '
+            'you supplied %s' % type(filename))
+
+    atom_indices = cast_indices(atom_indices)
+    
+    filename = str(filename)
+    with PDBTrajectoryFile(filename) as f:
+        atom_slice = slice(None) if atom_indices is None else atom_indices
+        if frame is not None:
+            coords = f.positions[[frame], atom_slice, :]
+        else:
+            coords = f.positions[::stride, atom_slice, :]
+        assert coords.ndim == 3, 'internal shape error'
+        n_frames = len(coords)
+
+        topology = f.topology
+        if atom_indices is not None:
+            topology = topology.subset(atom_indices)
+
+        if f.unitcell_angles is not None and f.unitcell_lengths is not None:
+            unitcell_lengths = np.array([f.unitcell_lengths] * n_frames)
+            unitcell_angles = np.array([f.unitcell_angles] * n_frames)
+        else:
+            unitcell_lengths = None
+            unitcell_angles = None
+
+        convert(coords, f.distance_unit, Trajectory._distance_unit, inplace=True)
+        convert(unitcell_lengths, f.distance_unit, Trajectory._distance_unit, inplace=True)
+
+    time = np.arange(len(coords))
+    if frame is not None:
+        time *= frame
+    elif stride is not None:
+        time *= stride
+
+    return Trajectory(xyz=coords, time=time, topology=topology,
+                      unitcell_lengths=unitcell_lengths,
+                      unitcell_angles=unitcell_angles)
 
 
+@_FormatRegistry.register_fileobject('.pdb')
 class PDBTrajectoryFile(object):
     """Interface for reading and writing Protein Data Bank (PDB) files
 
@@ -99,6 +181,7 @@ class PDBTrajectoryFile(object):
         self._topology = None
         self._positions = None
         self._mode = mode
+        self._last_topology = None
 
         if mode == 'r':
             PDBTrajectoryFile._loadNameReplacementTables()
@@ -144,11 +227,13 @@ class PDBTrajectoryFile(object):
             raise ValueError('Particle position is NaN')
         if np.any(np.isinf(positions)):
             raise ValueError('Particle position is infinite')
-
+        
+        self._last_topology = topology  # Hack to save the topology of the last frame written, allows us to output CONECT entries in write_footer()
+        
         atomIndex = 1
         posIndex = 0
         if modelIndex is not None:
-            print >> self._file, "MODEL     %4d" % modelIndex
+            print("MODEL     %4d" % modelIndex, file=self._file)
         for (chainIndex, chain) in enumerate(topology.chains):
             chainName = self._chain_names[chainIndex % len(self._chain_names)]
             residues = list(chain.residues)
@@ -165,20 +250,24 @@ class PDBTrajectoryFile(object):
                     else:
                         atomName = atom.name
                     coords = positions[posIndex]
-                    line = "ATOM  %5d %-4s %3s %s%4d    %s%s%s  1.00  0.00" % (
+                    if atom.element is not None:
+                        symbol = atom.element.symbol
+                    else:
+                        symbol = ' '
+                    line = "ATOM  %5d %-4s %3s %s%4d    %s%s%s  1.00  0.00          %2s  " % (
                         atomIndex % 100000, atomName, resName, chainName,
-                        (resIndex + 1) % 10000, _format_83(coords[0]),
-                        _format_83(coords[1]), _format_83(coords[2]))
-                    assert len(line) == 66, 'Fixed width overflow detected'
-                    print >> self._file, line
+                        (res.resSeq) % 10000, _format_83(coords[0]),
+                        _format_83(coords[1]), _format_83(coords[2]), symbol)
+                    assert len(line) == 80, 'Fixed width overflow detected'
+                    print(line, file=self._file)
                     posIndex += 1
                     atomIndex += 1
                 if resIndex == len(residues)-1:
-                    print >> self._file, "TER   %5d      %3s %s%4d" % (atomIndex, resName, chainName, resIndex+1)
+                    print("TER   %5d      %3s %s%4d" % (atomIndex, resName, chainName, resIndex+1), file=self._file)
                     atomIndex += 1
 
         if modelIndex is not None:
-            print >> self._file, "ENDMDL"
+            print("ENDMDL", file=self._file)
 
     def _write_header(self, unitcell_lengths, unitcell_angles):
         """Write out the header for a PDB file.
@@ -208,12 +297,63 @@ class PDBTrajectoryFile(object):
         box = list(unitcell_lengths) + list(unitcell_angles)
         assert len(box) == 6
 
-        print >>self._file, "CRYST1%9.3f%9.3f%9.3f%7.2f%7.2f%7.2f P 1           1 " % tuple(box)
+        print("CRYST1%9.3f%9.3f%9.3f%7.2f%7.2f%7.2f P 1           1 " % tuple(box), file=self._file)
 
     def _write_footer(self):
         if not self._mode == 'w':
             raise ValueError('file not opened for writing')
-        print >>self._file, "END"
+
+        # Identify bonds that should be listed as CONECT records.
+        standardResidues = ['ALA', 'ASN', 'CYS', 'GLU', 'HIS', 'LEU', 'MET', 'PRO', 'THR', 'TYR',
+                            'ARG', 'ASP', 'GLN', 'GLY', 'ILE', 'LYS', 'PHE', 'SER', 'TRP', 'VAL',
+                            'A', 'G', 'C', 'U', 'I', 'DA', 'DG', 'DC', 'DT', 'DI', 'HOH']
+        conectBonds = []
+        if self._last_topology is not None:
+            for atom1, atom2 in self._last_topology.bonds:
+                if atom1.residue.name not in standardResidues or atom2.residue.name not in standardResidues:
+                    conectBonds.append((atom1, atom2))
+                elif atom1.name == 'SG' and atom2.name == 'SG' and atom1.residue.name == 'CYS' and atom2.residue.name == 'CYS':
+                    conectBonds.append((atom1, atom2))
+        if len(conectBonds) > 0:
+            
+            # Work out the index used in the PDB file for each atom.
+            
+            atomIndex = {}
+            nextAtomIndex = 0
+            prevChain = None
+            for chain in self._last_topology.chains:
+                for atom in chain.atoms:
+                    if atom.residue.chain != prevChain:
+                        nextAtomIndex += 1
+                        prevChain = atom.residue.chain
+                    atomIndex[atom] = nextAtomIndex
+                    nextAtomIndex += 1
+            
+            # Record which other atoms each atom is bonded to.
+            
+            atomBonds = {}
+            for atom1, atom2 in conectBonds:
+                index1 = atomIndex[atom1]
+                index2 = atomIndex[atom2]
+                if index1 not in atomBonds:
+                    atomBonds[index1] = []
+                if index2 not in atomBonds:
+                    atomBonds[index2] = []
+                atomBonds[index1].append(index2)
+                atomBonds[index2].append(index1)
+            
+            # Write the CONECT records.
+            
+            for index1 in sorted(atomBonds):
+                bonded = atomBonds[index1]
+                while len(bonded) > 4:
+                    print("CONECT%5d%5d%5d%5d" % (index1, bonded[0], bonded[1], bonded[2]), file=self._file)
+                    del bonded[:4]
+                line = "CONECT%5d" % index1
+                for index2 in bonded:
+                    line = "%s%5d" % (line, index2)
+                print(line, file=self._file)
+        print("END", file=self._file)
         self._footer_written = True
 
     @classmethod
@@ -231,7 +371,7 @@ class PDBTrajectoryFile(object):
             cycle through to choose chain names.
         """
         for item in values:
-            if not isinstance(item, basestring) and len(item) == 1:
+            if not isinstance(item, str) and len(item) == 1:
                 raise TypeError('Names must be a single character string')
         cls._chain_names = values
 
@@ -285,7 +425,7 @@ class PDBTrajectoryFile(object):
                 resName = residue.get_name()
                 if resName in PDBTrajectoryFile._residueNameReplacements:
                     resName = PDBTrajectoryFile._residueNameReplacements[resName]
-                r = self._topology.add_residue(resName, c)
+                r = self._topology.add_residue(resName, c, residue.number)
                 if resName in PDBTrajectoryFile._atomNameReplacements:
                     atomReplacements = PDBTrajectoryFile._atomNameReplacements[resName]
                 else:
@@ -408,7 +548,7 @@ class PDBTrajectoryFile(object):
             except KeyError:
                 try:
                     symbol = atom_name[0:2].strip().rstrip("AB0123456789").lstrip("0123456789")
-                    element = elem.get_by_symbol(symbol)                                    
+                    element = elem.get_by_symbol(symbol)
                 except KeyError:
                     element = None
 
@@ -430,6 +570,14 @@ class PDBTrajectoryFile(object):
     def __exit__(self, *exc_info):
         self.close()
 
+    def __len__(self):
+        "Number of frames in the file"
+        if str(self._mode) != 'r':
+            raise NotImplementedError('len() only available in mode="r" currently')
+        if not self._open:
+            raise ValueError('I/O operation on closed file')
+        return len(self._positions)
+
 
 def _format_83(f):
     """Format a single float into a string of width 8, with ideally 3 decimal
